#include <cmath>
#include <fstream>
#include <iostream>
#include <vector>

#include <H5Cpp.h>

/**
 * A simple enum to indicate if we are on a "red" or "blue" row
 */
enum RowType {
    RedRow,
    BlueRow
};

/**
 * A simple enum to track the filter type
 */
enum FilterType {
    RedFilter   = 0,
    GreenFilter = 1,
    BlueFilter  = 2
};

/**
 * This is a simple program that generates a 320 x 240, Bayer pattern
 * pixel database file for use by the DIRSIG4 data-driven detector
 * array model.  A pixel database file includes the pixel line-of-sight
 * (LOS) vectors (as an X/Y angle pair) and optional spectral response
 * information.  The program creates the pixel databse using a minimal
 * set of inputs including the focal length, pixel pitch and number of
 * pixels in each dimension.  The Bayer pattern is imposed by the
 * assignment of spectral response functions to each pixel via a channel
 * index.
 * 
 * This program can be easily modified to model a different size array,
 * to use a different focal length, or to use a different color filter
 * array (CFA) pattern (other than Bayer).
 */
int
main( int argCount, char ** argVector )
{
    // the focal length is 100 mm
    const double focalLength = 100.0e-03;

    // the array size is 320 x 240
    const int xArraySize = 320;
    const int yArraySize = 240;

    // the pixel pitch is 40 microns
    const double pixelPitch = 40.0e-06;

    // the pixel IFOV can be computed
    const double pixelIfov = atan( pixelPitch / focalLength );

    // try to open the HDF file
    H5::H5File* hdfFile = 0;
    try {
        hdfFile = new H5::H5File( "pixels.hdf", H5F_ACC_TRUNC );
    }
    catch( ... ) {
        std::cerr << "Unexpected exception in opening HDF file!" << std::endl;
        return 1;
    }

    // allocate buffers for the geometry and channel tables
    size_t pixelCount = xArraySize * yArraySize;
    const hsize_t geomTableWidth = 2;
    const hsize_t channelDataWidth = 1;
    std::vector< double > angleData( pixelCount * geomTableWidth );
    std::vector< double > ifovData( pixelCount * geomTableWidth );
    std::vector< unsigned int > channelData( pixelCount * channelDataWidth );

    // iterate through all the Y rows in this array
    for( int yy = 0; yy < yArraySize; yy++ ) {

        // figure out if this "red" or "blue" row
        RowType rowType;
        if(( yy % 2 ) == 0 ) {
            rowType = RedRow;
        }
        else {
            rowType = BlueRow;
        }

        // iterate through all the X pixels in this row
        for( int xx = 0; xx < xArraySize; xx++ ) {

            // compute the pixel X/Y location
            double xLoc = ( xx - ( xArraySize / 2.0 )) * pixelPitch;
            double yLoc = ( yy - ( yArraySize / 2.0 )) * pixelPitch;

            // compute the pixel X/Y angles
            double xAngle = atan( xLoc / focalLength );
            double yAngle = atan( yLoc / focalLength );

            // figure out what filter it gets
            FilterType filterType;
            if( rowType == RedRow ) {
                if(( xx % 2 ) == 0 ) {
                    filterType = RedFilter;
                }
                else {
                    filterType = GreenFilter;
                }
            }
            else if( rowType == BlueRow ) {
                if(( xx % 2 ) == 1 ) {
                    filterType = BlueFilter;
                }
                else {
                    filterType = GreenFilter;
                }
            }

            // write the pixel geometry values to the geometry table
            angleData.at((( yy * xArraySize ) + xx ) * 2 + 0 ) = xAngle;
            angleData.at((( yy * xArraySize ) + xx ) * 2 + 1 ) = yAngle;
            ifovData.at((( yy * xArraySize ) + xx ) * 2 + 0 ) = pixelIfov;
            ifovData.at((( yy * xArraySize ) + xx ) * 2 + 1 ) = pixelIfov;

            // write the pixel channel to the channel data
            channelData.at(( yy * xArraySize ) + xx ) =
                    static_cast< unsigned int >( filterType );
        }
    }

    hsize_t dataRank;
    hsize_t dataDims[ 2 ];
    hsize_t chunkDims[ 2 ];
    H5::DataSpace dataSpace;
    hsize_t zLibChunkSize = 8192;
    int zLibFactor = 1;

    // store the pixel geometry table
    dataRank = 2;
    dataDims[ 0 ] = pixelCount;
    dataDims[ 1 ] = geomTableWidth;
    chunkDims[ 0 ] = zLibChunkSize;
    chunkDims[ 1 ] = geomTableWidth;
    dataSpace = H5::DataSpace( dataRank, dataDims );
    H5::DSetCreatPropList angleDataPlist;
    angleDataPlist.setChunk( dataRank, chunkDims );
    angleDataPlist.setDeflate( zLibFactor );
    H5::DataSet angleDataSet = hdfFile->createDataSet( "/AngleTable",
            H5::PredType::NATIVE_DOUBLE, dataSpace, angleDataPlist );
    angleDataSet.write( angleData.data(), H5::PredType::NATIVE_DOUBLE );
    angleDataSet.close();

    dataRank = 2;
    dataDims[ 0 ] = pixelCount;
    dataDims[ 1 ] = geomTableWidth;
    chunkDims[ 0 ] = zLibChunkSize;
    chunkDims[ 1 ] = geomTableWidth;
    dataSpace = H5::DataSpace( dataRank, dataDims );
    H5::DSetCreatPropList ifovDataPlist;
    ifovDataPlist.setChunk( dataRank, chunkDims );
    ifovDataPlist.setDeflate( zLibFactor );
    H5::DataSet ifovDataSet = hdfFile->createDataSet( "/IfovTable",
            H5::PredType::NATIVE_DOUBLE, dataSpace, ifovDataPlist );
    ifovDataSet.write( ifovData.data(), H5::PredType::NATIVE_DOUBLE );
    ifovDataSet.close();

    // store the pixel channel table
    dataRank = 2;
    dataDims[ 0 ] = pixelCount;
    dataDims[ 1 ] = channelDataWidth;
    chunkDims[ 0 ] = zLibChunkSize;
    chunkDims[ 1 ] = channelDataWidth;
    dataSpace = H5::DataSpace( dataRank, dataDims );
    H5::DSetCreatPropList channelDataPlist;
    channelDataPlist.setChunk( dataRank, chunkDims );
    channelDataPlist.setDeflate( zLibFactor );
    H5::DataSet channelDataSet = hdfFile->createDataSet( "/ChannelTable",
            H5::PredType::NATIVE_UINT, dataSpace, channelDataPlist );
    channelDataSet.write( channelData.data(), H5::PredType::NATIVE_UINT );
    channelDataSet.close();

    hdfFile->close();

}
